/*
 * A system pager implementation residing in the same task as a program.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/re/c/dataspace.h>
#include <l4/util/util.h>

#include <mem/memory_utils.h>
#include <systypes/base.h>

#include <stdio.h>

#include "dataspace_client.h"
#include "internal_pager.h"



#define DEBUG 0



/* A simple system pager also acting as a region mapper. */

InternalPager::InternalPager(address_t start, address_t end)
: ExecPager(start, end)
{
}

/* Handle a general exception. */

long InternalPager::exception(l4_exc_regs_t regs, l4_snd_fpage_t *region)
{
  (void) region;

  printf("InternalPager::exception(...) -> pfa = %lx, pc = %lx\n", l4_utcb_exc_pfa(&regs), l4_utcb_exc_pc(&regs));

#if DEBUG
  while (1)
    l4_sleep_forever();
#endif

  return L4_EOK;
}

/* Handle a page fault using any configured regions. */

long InternalPager::page_fault(l4_umword_t pfa, l4_umword_t pc, l4_snd_fpage_t *region)
{
  l4_umword_t addr = pfa & ~7UL, flags = pfa & 7;

#if DEBUG
  printf("page_fault(%lx, %lx) -> %lx (%lx) -> ", pfa, pc, addr, flags);
  l4_thread_regs_t *regs = l4_utcb_tcr();
  printf("regs = %p\n; regs->user[0] = %lx\n", regs, regs->user[0]);
#endif

  /* Find the first region whose start address is beyond the fault address,
     testing if any immediately preceding region contains the fault address. */

  MappedRegions::iterator it = _regions.upper_bound(addr);

  if (it != _regions.begin())
    it--;
  else
  {
    printf("not mapped at %lx for pc %lx\n", addr, pc);
    return -L4_ENOMEM;
  }

  MappedRegion &r = it->second;

  if ((addr >= r.start) && (addr < r.start + r.size))
  {
    address_t window_size = L4_PAGESIZE;
    address_t window_base = trunc(addr, window_size);
    address_t offset = addr - r.start;
    address_t page_addr = trunc(addr, L4_PAGESIZE);
    address_t hot_spot = page_addr - window_base;

    /* Interact with the region's dataspace, specifying a receive window for a
       map operation. Here, a single page is specified. */

    client_Dataspace dataspace(r.ds);
    l4_snd_fpage_t rw_region = {0, l4_fpage(window_base, L4_PAGESHIFT, 0)};
    map_flags_t map_flags = map_flags_for_fault(flags);

#if DEBUG
    printf("window_base = %lx; window_size = %lx\n", window_base, window_size);
    printf("region = {%lx, {%lx, %d}}\n", rw_region.snd_base, l4_fpage_memaddr(rw_region.fpage), l4_fpage_size(rw_region.fpage));
    printf("map(%lx, %lx, %lx) -> %lx\n", offset, hot_spot, map_flags, r.ds);
#endif

    long err = dataspace.map(offset, hot_spot, map_flags & r.flags, &rw_region);

    /* Indicate an unspecified result, since the mapping should have taken
       place. */

    *region = {0, l4_fpage_invalid()};

    return err;
  }

#if DEBUG
  printf("not mapped!\n");
#endif

  return -L4_ENOMEM;
}

/* Attach a region for provision when page faults occur. This is required in
   the initialisation of a program by the C library which requires a region
   mapper. */

long InternalPager::attach(address_t *start, address_t size, map_flags_t flags,
                           l4_cap_idx_t ds, address_t offset,
                           unsigned char align)
{
  long err = ExecPager::find(start, &size, flags, offset, align);

  if (!err)
  {
    MappedRegion r(*start, size, flags & L4RE_DS_F_RIGHTS_MASK, ds);
    add(r);
  }

  return err;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
